#!/usr/local/bin/perl

# Copyright (C) 2010-2012 Trizen <echo dHJpemVueEBnbWFpbC5jb20K | base64 -d>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#-------------------------------------------------------
#  GTK Youtube Viewer
#  Created on: 12 September 2010
#  Latest edit on: 09 May 2012
#  Website: http://trizen.googlecode.com
#-------------------------------------------------------

eval 'exec perl $0 ${1+"$@"}'
  if 0;    # not running under some shell

use strict;

#-----FOR DEBUG ONLY----#
# use diagnostics -v;
# use warnings FATAL => 'all';
#-----------------------#

use 5.010;
use Gtk2 qw(-init);
use autouse qw(XML::Fast xml2hash);
use autouse qw(URI::Escape uri_escape_utf8);
use File::Spec::Functions qw(path tmpdir curdir rel2abs);

# Application modules
require './lib/Config.pm';

my $appname  = 'GTK Youtube Viewer';
my $version  = '2.0.5';
my $execname = 'gtk-youtube-viewer';

# Configuration dir/file
my $config_dir = (
                  exists $ENV{XDG_CONFIG_HOME}
                  ? $ENV{XDG_CONFIG_HOME}
                  : ($ENV{HOME} || $ENV{LOGDIR} || (getpwuid($<))[7] || `echo -n ~`) . '/.config'
                 ) . "/$execname";

my $config_file = "$config_dir/$execname.conf";

# Unchangeable variables goes here
my %constant = (
                gdata_version     => 2,
                comments_per_page => 25,
                win32             => $^O eq 'MSWin32' ? 1 : 0,
               );

# Set $PATH to @path
my @path = path();

# CLI Youtube Viewer
my $cli_youtube_viewer =
  -e '/usr/bin/youtube-viewer'
  ? '/usr/bin/youtube-viewer'
  : rel2abs('./youtube-viewer');

# Locating gcap
my $gcap;
foreach my $path (@path, @INC) {
    if (-e "$path/gcap") {
        $gcap = "$path/gcap";
        last;
    }
}

# Get mplayer line
sub get_mplayer {
    if ($constant{'win32'}) {
        my $smplayer = $ENV{'ProgramFiles'} . '\\SMPlayer\\mplayer\\mplayer.exe';
        if (-e $smplayer) {
            return $smplayer;    # Windows MPlayer
        }
        else {
            warn "\n\n!!! Install SMPlayer in order to stream YouTube videos.\n\n";
            return $smplayer;
        }
    }
    else {
        return 'mplayer'         # *NIX MPlayer
    }
}

# Main configuration
my %CONFIG = (

    # MPlayer options
    cache                => 30000,
    cache_min            => 5,
    lower_cache          => 1000,
    lower_cache_min      => 3,
    mplayer              => get_mplayer(),
    mplayer_srt_settings => '-unicode -utf8',
    mplayer_arguments    => '-prefer-ipv4 -really-quiet -cache %d -cache-min %d',

    # GUI options
    clean_text_entries_on_click => 1,
    show_thumbs                 => 1,
    clear_list                  => 1,
    default_notebook_page       => 1,
    mainw_size                  => '700x400',

    # Youtube options
    start_index      => 1,
    start_with_page  => 1,
    results          => 10,
    resolution       => 1080,
    query_parameters => undef,
    time_sort        => 'all_time',
    order_by         => 'relevance',

    # URI options
    thumb_url           => 'http://i2.ytimg.com/vi/%s/%s.jpg',
    youtube_video_url   => 'http://www.youtube.com/watch?v=%s',
    youtube_profile_url => 'http://www.youtube.com/user/%s',
    feeds_main_url      => 'http://gdata.youtube.com/feeds/api',
    get_video_info      => 'http://www.youtube.com/get_video_info',
    playlists_url       => 'http://www.youtube.com/playlist?list=%s',
    google_client_login => 'https://www.google.com/accounts/ClientLogin',

    # Subtitle options
    srt_language => 'en',
    tmp_dir      => tmpdir(),
    gcap         => $gcap,

    # Others
    debug                    => 0,
    fullscreen               => 0,
    use_lower_cache          => 0,
    use_threads              => 0,
    use_threads_for_thumbs   => 0,
    show_comments_by_default => 0,
    lwp_timeout              => 60,
    hpaned_position          => 420,
    thousand_separator       => q{,},
    perl_bin                 => $^X,
    default_thumb            => 'default',
    downloads_folder         => $ENV{'HOME'},
    browser                  => undef,
    terminal                 => find_terminal(),
    cli_youtube_viewer       => $cli_youtube_viewer,
    users_list               => "$config_dir/youtube_users.txt",
);

my %MPLAYER;

# MPlayer variable arguments
sub set_mplayer_arguments {
    my ($cache, $cache_min) = @_;
    $MPLAYER{mplayer_arguments} = sprintf $CONFIG{mplayer_arguments}, $cache, $cache_min;
    $MPLAYER{fullscreen} = $CONFIG{fullscreen} ? '-fs' : '';
    return 1;
}

# Save hash config to file
sub write_config_to_file {
    Config::save_hash($config_file, \%CONFIG);
}

# Creating config unless it exists
unless (-e $config_file) {
    require File::Path;
    File::Path::make_path($config_dir);
    write_config_to_file();
}

# SIG handlers
foreach my $sig (qw(HUP TERM INT KILL)) {
    $SIG{$sig} = \&on_mainw_destroy;
}

# Locate glade interface
my ($interface_path, $icons_path);
if (-e "$execname.glade") {
    $interface_path = "$execname.glade";
    $icons_path     = 'icons';
}
else {
    $interface_path = "/usr/share/$execname/$execname.glade";
    $icons_path     = "/usr/share/$execname/icons";
}

my $share_path = $icons_path;
$share_path =~ s/icons$//;

# Defining GUI
my $gui = 'Gtk2::Builder'->new;
$gui->add_from_file($interface_path);
$gui->connect_signals(undef);

# -------------  Get GUI objects ------------- #
my (
    $mainw,               $users_list_window,         $help_window,          $prefernces_window,
    $login_to_youtube,    $errors_window,             $details_window,       $about_window,
    $treeview,            $liststore,                 $config_view,          $statusbar,
    $thumbs_column,       $textview_help,             $sort_combobox,        $sort_by_time_combobox,
    $resolution_combobox, $spin_results,              $users_treeview,       $users_liststore,
    $thumbs_checkbutton,  $videos_playlists_combobox, $spin_start_with_page, $fullscreen_checkbutton,
    $notebook,            $search_entry,              $from_author_entry,    $gif_spinner,
    $hbox2,               $feeds_window,              $feeds_treeview,       $feeds_liststore,
    $feeds_statusbar,     $safesearch_combobox,       $hd_combobox,          $caption_combobox,
    $duration_combobox,
   );

sub get_objects {
    my %objects = (

        # Windows
        '__MAIN__'          => \$mainw,
        'users_list_window' => \$users_list_window,
        'help_window'       => \$help_window,
        'prefernces_window' => \$prefernces_window,
        'errors_window'     => \$errors_window,
        'login_to_youtube'  => \$login_to_youtube,
        'details_window'    => \$details_window,
        'aboutdialog1'      => \$about_window,
        'feeds_window'      => \$feeds_window,

        # Others
        'treeview1'                 => \$users_treeview,
        'feeds_statusbar'           => \$feeds_statusbar,
        'treeview2'                 => \$treeview,
        'feeds_treeview'            => \$feeds_treeview,
        'liststore1'                => \$liststore,
        'liststore2'                => \$users_liststore,
        'liststore11'               => \$feeds_liststore,
        'textview3'                 => \$config_view,
        'search_entry'              => \$search_entry,
        'statusbar1'                => \$statusbar,
        'treeviewcolumn2'           => \$thumbs_column,
        'textview2'                 => \$textview_help,
        'from_author_entry'         => \$from_author_entry,
        'notebook1'                 => \$notebook,
        'combobox1'                 => \$sort_combobox,
        'combobox2'                 => \$sort_by_time_combobox,
        'combobox3'                 => \$resolution_combobox,
        'combobox4'                 => \$duration_combobox,
        'combobox5'                 => \$caption_combobox,
        'combobox6'                 => \$hd_combobox,
        'combobox7'                 => \$safesearch_combobox,
        'spinbutton1'               => \$spin_results,
        'spinbutton2'               => \$spin_start_with_page,
        'thumbs_checkbutton'        => \$thumbs_checkbutton,
        'videos_playlists_combobox' => \$videos_playlists_combobox,
        'fullscreen_checkbutton'    => \$fullscreen_checkbutton,
        'gif_spinner'               => \$gif_spinner,
        'hbox2'                     => \$hbox2,
    );

    while (my ($key, $value) = each %objects) {
        ${$value} = $gui->get_object($key);
    }
}
get_objects();

#---------------------- LOAD IMAGES ----------------------#
my $app_icon_pixbuf  = 'Gtk2::Gdk::Pixbuf'->new_from_file("$icons_path/$execname.png");
my $user_icon_pixbuf = 'Gtk2::Gdk::Pixbuf'->new_from_file_at_size("$icons_path/user.png", 16, 16);
my $feed_icon_pixbuf = 'Gtk2::Gdk::Pixbuf'->new_from_file_at_size("$icons_path/feed_icon.png", 16, 16);
my $default_thumb    = 'Gtk2::Gdk::Pixbuf'->new_from_file_at_size("$icons_path/default_thumb.jpg", 120, 90);

# Setting application title and icon
$mainw->set_title("$appname $version");
$mainw->set_icon($app_icon_pixbuf);

# For YouTube
my $key = 'AI39si5xZtotT-ABtXEHNYpPnfez4T9hfNTkMlWti5gVCbFOZ-wyw70RTTguH_53klpmj3G98sTJGXJF5YY61Zcu1r5XmR2w3Q';
my %lwp_header = ('X-GData-Key' => "key=$key");

#---------------------- YOUTUBE STUFF ----------------------#
# itag => resolution
my %itags = (
             37 => 1080,
             22 => 720,
             35 => 480,
             43 => 210,
             34 => 360,
             5  => 240
            );

#---------------------- GLOBAL VARIABLES ----------------------#
my ($youtube_gdata_url, @results, %streaming);    # other global variables

#------------------------ REGEXP AREA ------------------------#
my $valid_playlist_code = qr/^(?:PL)?([0-9A-Z]{16})$/;
my $get_youtube_code    = qr{\b(?>v|embed|youtu[.]be)(?>[=/]|%3D)([\w\-]{11})};

# $1 will be the playlist code
my $get_playlist_code = qr{(?:(?:(?>playlist\?list|view_play_list\?p)=)|\w#p/c/)(?:PL)?([A-Z0-9]{16})};

# $1 will be the Youtube username
my $get_username = qr{^https?://(?:www[.])?youtube[.]com/(?:user/)?(\w{2,})(?:[?].*)?$};

# The bellow regex will validate an HTTP url.
# If it is valid, we will try to get an youtube
# video code from that website using /$get_youtube_code/
my $valid_url = qr{^
        ################# This regex will validate an HTTP URL #################

                https?://                # http or https followed by ://
                [[:alnum:]]              # first character must be a-zA-Z0-9
                (?:(?:(?:\w*-+\w+|\w+)*  # words, dash, words OR only words
                \.(?=\w))+?              # point if followed by word char
                |                        # OR (validates http://x.yz)
                \.)                      # a single dot
                \w{2,6}                  # domain (words between 2 and 6 chars)
                (?:[#/?!]                # characters after domain
                [#-)+-;=?\\~\w]*)*       # the rest characters of the string
$}x;

#---------------------- LWP::UserAgent ----------------------#
my $lwp;

sub set_lwp_useragent {
    require LWP;
    require HTTP::Request;

    warn "* Setting LWP::UserAgent called from line: ${\caller}\n" if $CONFIG{debug};
    $lwp = 'LWP::UserAgent'->new(
                                 keep_alive    => 1,
                                 env_proxy     => 1,
                                 show_progress => 1,
                                 timeout       => $CONFIG{lwp_timeout},
                                 agent         => 'Mozilla/5.0 (X11; U; Linux i686; en-US) Chrome/10.0.648.45',
                                );
    push @{$lwp->requests_redirectable}, 'POST';
}

sub lwp_get {
    set_lwp_useragent() unless exists $INC{'LWP.pm'};
    return $lwp->get($_[0], %lwp_header)->content;
}

# ---------------- Threads ---------------- #
my ($queue, $jobs);

sub set_threads {
    return 1 if defined $queue;
    my $animation = 'Gtk2::Gdk::PixbufAnimation'->new_from_file("$icons_path/spinner.gif");
    $gif_spinner->set_from_animation($animation);
    warn "* Initializing threads...\n";

    require threads;
    require Thread::Queue;

    *lwp_get = *threads_lwp_get;
    $queue   = 'Thread::Queue'->new;
    $jobs    = 'Thread::Queue'->new;
    threads->create(
        sub {
            set_lwp_useragent() unless exists $INC{'LWP.pm'};
            while (defined(my $url = $jobs->dequeue)) {
                $queue->enqueue($lwp->get($url, %lwp_header)->content);
            }
        }
    )->detach();
}

sub threads_lwp_get {
    set_lwp_useragent() unless exists $INC{'LWP.pm'};
    set_threads() unless defined $queue;
    $gif_spinner->show;
    my ($url) = @_;
    $jobs->enqueue($url);
    while ($queue->pending == 0) {
        'Gtk2'->main_iteration;
        if (defined(my $lwp_result = $queue->dequeue_nb)) {
            $gif_spinner->hide;
            return $lwp_result;
        }
    }
    undef $queue;
    undef $jobs;
}

# Set config file to %CONFIG
sub apply_configuration {
    my $config_ref = do $config_file;
    if (ref $config_ref eq 'HASH') {
        while (my ($key, $value) = each %$config_ref) {
            $CONFIG{$key} = $value;
        }
    }
    else {
        warn "\n[!] Configuration file contains some errors!\n",
          "[*] Trying to regenerate the configuration file...\n",
          (write_config_to_file() ? "[*] Done!\n" : "[!] Unable to regenerate $config_file: $!\n");
    }

    # Menu items
    $fullscreen_checkbutton->set_active($CONFIG{fullscreen});

    # Spin button results setting config value
    $spin_results->set_value($CONFIG{results});

    # Spin button start with page
    $spin_start_with_page->set_value($CONFIG{start_with_page});

    # Checking thumbs button
    $thumbs_checkbutton->set_active($CONFIG{show_thumbs});

    # Auto-login
    if (exists $CONFIG{auth} and $CONFIG{auth}) {
        set_auth_key($CONFIG{auth});
    }
    else {
        $statusbar->push(1, 'Not logged');
    }

    # Setup threads
    if ($CONFIG{use_threads}) {
        set_threads();
    }

    # Combo boxes setting config value
    if (exists $CONFIG{active_sort_by_combobox}) {
        $sort_combobox->set_active($CONFIG{active_sort_by_combobox});
    }
    if (exists $CONFIG{active_time_sort_by_combobox}) {
        $sort_by_time_combobox->set_active($CONFIG{active_time_sort_by_combobox});
    }
    if (exists $CONFIG{active_resolution_combobox}) {
        $resolution_combobox->set_active($CONFIG{active_resolution_combobox});
    }
    if (exists $CONFIG{active_videos_playlists}) {
        $videos_playlists_combobox->set_active($CONFIG{active_videos_playlists});
    }
    if (exists $CONFIG{active_hd_combobox}) {
        $hd_combobox->set_active($CONFIG{active_hd_combobox});
    }
    if (exists $CONFIG{active_caption_combobox}) {
        $caption_combobox->set_active($CONFIG{active_caption_combobox});
    }
    if (exists $CONFIG{active_duration_combobox}) {
        $duration_combobox->set_active($CONFIG{active_duration_combobox});
    }
    if (exists $CONFIG{active_safe_search_combobox}) {
        $safesearch_combobox->set_active($CONFIG{active_safe_search_combobox});
    }

    # Resize the main window
    $mainw->set_default_size(split(/x/i, $CONFIG{mainw_size}, 2));
    $mainw->reshow_with_initial_size;

    # Set HPaned position
    $hbox2->set_position($CONFIG{hpaned_position});

    # Select text from text entry
    $search_entry->select_region(0, -1);
}

# Apply the configuration file
apply_configuration();

# YouTube usernames
my %users_table = map { lc($_) => $_ } (
                                        'AtGoogleTalks',  'AtheistsAreSkeptics', 'Best0fScience',        'DiscoveryTV',
                                        'Google',         'GoogleChrome',        'GoogleTechTalks',      'KhanAcademy',
                                        'NASAtelevision', 'NatCen4ScienceEd',    'GoogleDevelopers',     'RichardDawkinsDotNet',
                                        'ScienceChannel', 'SpaceRip',            'AnimalPlanetTV',       'TEDtalksDirector',
                                        'MIT',            'SixtySymbols',        'SpitzerScienceCenter', 'EvolutionDocumentary',
                                        'FOSDEMtalks'
                                       );
set_usernames();

# Verifies if https is available
if (my @https_url_keys = grep { defined $CONFIG{$_} and $CONFIG{$_} =~ m{^https://} } keys %CONFIG) {
    unless (grep { -e "$_/LWP/Protocol/https.pm" } @INC) {
        foreach my $https_url_key (@https_url_keys) {
            $CONFIG{$https_url_key} =~ s{^https://}{http://};
        }
        warn "[!] Switched to HTTP protocol instead of HTTPS!\n",
          "[!] In order to use the HTTPS protocol, please install the 'LWP::Protocol::https' module\n";
    }
}

# Set text to a 'textview' object
sub set_text {
    my ($object, $text) = @_;
    my $object_buffer = $object->get_buffer;
    $object_buffer->set_text($text);
    $object->set_buffer($object_buffer);
    return 1;
}

# Get text from a 'textview' object
sub get_text {
    my ($object)      = @_;
    my $object_buffer = $object->get_buffer;
    my $start_iter    = $object_buffer->get_start_iter;
    my $end_iter      = $object_buffer->get_end_iter;
    return $object_buffer->get_text($start_iter, $end_iter, undef);
}

# __DIE__ handle
$SIG{__DIE__} = sub {
    my $error = join('', @_);
    set_text(
        $gui->get_object('errors_text'),
        $error . do {
            if ($error =~ m{^Can't locate (.+?)\.pm\b}) {
                my $module = $1;
                $module =~ s{[/\\]+}{::}g;
                "\nThe module $module is required!\n\nTo install the module, just type in terminal:\n\tsudo cpan $module\n";
            }
          }
    );
    $errors_window->show;
    return 1;
};

# Setting application icons
my $user_icon = $gui->get_object('icon_from_pixbux')->new_from_pixbuf($user_icon_pixbuf);
$gui->get_object('username_list')->set_image($user_icon);
my $feed_icon_image = $gui->get_object('icon_from_pixbux')->new_from_pixbuf($feed_icon_pixbuf);
$gui->get_object('button6')->set_image($feed_icon_image);
my $subscribe_icon = $gui->get_object('icon_from_pixbux')->new_from_pixbuf($feed_icon_pixbuf);
$gui->get_object('button23')->set_image($subscribe_icon);

# Treeview signals
$treeview->signal_connect('button_press_event', \&menu_popup);
$users_treeview->signal_connect('button_press_event', \&users_menu_popup);

# Menu popup
sub menu_popup {
    my ($treeview, $event) = @_;
    if ($event->button != 3) {
        return 0;
    }
    my $menu = $gui->get_object('detailsmenu');
    $menu->popup(undef, undef, undef, undef, $event->button, $event->time);
    return 0;
}

sub users_menu_popup {
    my ($treeview, $event) = @_;
    if ($event->button != 3) {
        return 0;
    }
    my $menu = $gui->get_object('user_option_menu');
    $menu->popup(undef, undef, undef, undef, $event->button, $event->time);
    return 0;
}

# Setting help text
set_text(
    $textview_help, <<"HELP_TEXT"
* Links
    main website: http://trizen.googlecode.com
    development website: https://github.com/trizen/$execname
    developer's website: http://trizen.go.ro\n
* Developer
    Trizen <trizenx\@gmail.com>\n
* Contributor
    Ovidiu D. Ni\x{21b}an <nitanovidiu\@gmail.com>\n
* Config file
    $config_file\n
* Users list
    $CONFIG{users_list}\n
* Key binds\n
-Main window
CTRL+H : help window
CTRL+L : login window
CTRL+P : preferences window
CTRL+U : username list window
CTRL+Y : cli youtube viewer
CTRL+D : video details window
CTRL+N : next page of results
CTRL+F : show feeds window
CTRL+R : show related videos of a selected video
CTRL+M : show more videos from the author of a selected video
CTRL+S : save uploader, from a selected video, to favorites
CTRL+Q : quit the application
F11 : minimize-maximize the main window\n
-Preferences window
CTRL+S : save the configuration\n
-Other windows
ESC : to close that window\n
* Configuration\n
use_threads
    1 to use threads\n
cache & cache_min
    are used for resolutions higher than 480p\n
lower_cache & lower_cache_min
    are used for resolutions lower than 720p\n
use_lower_cache
    1 to always use the lower_cache (for slow connections)\n
show_thumbs
    1 to show YouTube video thumbnails (may be slow)\n
clear_list
    1 to clean the treeview list before each search\n
gcap
    path to gcap application ( see http://gcap.googlecode.com )\n
srt_language
    default subtitle language (when gcap is available)\n
tmp_dir
    is used to store the closed captions from YouTube (.srt files)\n
query_parameters
    any of the available parameters (example: '&duration=long&caption')
    http://code.google.com/intl/ro/apis/youtube/2.0/developers_guide_protocol_api_query_parameters.html\n
HELP_TEXT
        );

# ------------------- Accels ------------------- #

# Main window
my $accel = Gtk2::AccelGroup->new;
$accel->connect(ord('h'), ['control-mask'], ['visible'], \&show_help_window);
$accel->connect(ord('l'), ['control-mask'], ['visible'], \&show_login_to_youtube_window);
$accel->connect(ord('p'), ['control-mask'], ['visible'], \&show_preferences_window);
$accel->connect(ord('q'), ['control-mask'], ['visible'], \&on_mainw_destroy);
$accel->connect(ord('u'), ['control-mask'], ['visible'], \&show_users_list_window);
$accel->connect(ord('y'), ['control-mask'], ['visible'], \&run_cli_youtube_viewer);
$accel->connect(ord('d'), ['control-mask'], ['visible'], \&show_details_window);
$accel->connect(ord('n'), ['control-mask'], ['visible'], \&next_page);
$accel->connect(ord('f'), ['control-mask'], ['visible'], \&show_feeds_window);
$accel->connect(ord('s'), ['control-mask'], ['visible'], \&add_user_to_favorites);
$accel->connect(ord('r'), ['control-mask'], ['visible'], \&show_related_videos);
$accel->connect(ord('m'), ['control-mask'], ['visible'], \&show_more_videos_from_username);
$accel->connect(0xffc8,   ['lock-mask'],    ['visible'], \&maximize_unmaximize_mainw);
$mainw->add_accel_group($accel);

# Other windows (ESC key to close them)
$accel = Gtk2::AccelGroup->new;
$accel->connect(0xff1b, ['lock-mask'], ['visible'], \&hide_users_list_window);
$users_list_window->add_accel_group($accel);

$accel = Gtk2::AccelGroup->new;
$accel->connect(0xff1b, ['lock-mask'], ['visible'], \&hide_feeds_window);
$feeds_window->add_accel_group($accel);

$accel = Gtk2::AccelGroup->new;
$accel->connect(0xff1b,   ['lock-mask'],    ['visible'], \&hide_preferences_window);
$accel->connect(ord('s'), ['control-mask'], ['visible'], \&save_configuration);
$prefernces_window->add_accel_group($accel);

$accel = Gtk2::AccelGroup->new;
$accel->connect(0xff1b, ['lock-mask'], ['visible'], \&hide_help_window);
$help_window->add_accel_group($accel);

$accel = Gtk2::AccelGroup->new;
$accel->connect(0xff1b, ['lock-mask'], ['visible'], \&hide_details_window);
$details_window->add_accel_group($accel);

# ------------------ Authentication ------------------ #

sub set_auth_key {
    my ($auth) = @_;
    $lwp_header{'Authorization'} = "GoogleLogin auth=$auth";
    change_subscription_page(1);
    $statusbar->push(1, "Logged...");
    $gui->get_object('login_check_button')->set_active(1);
}

sub authenticate {
    my $email    = $gui->get_object('email')->get_text;
    my $password = $gui->get_object('password')->get_text;
    set_lwp_useragent() unless exists $INC{'LWP.pm'};
    my $resp = $lwp->post(
                          $CONFIG{google_client_login},
                          [Content => 'application/x-www-form-urlencoded',
                           Email   => $email,
                           Passwd  => $password,
                           service => 'youtube',
                           source  => "$appname $version"
                          ]
                         );
    hide_login_to_youtube_window();
    if ($resp->{_content} =~ /^Auth=(.+)/m) {
        my $auth = $1;
        $CONFIG{email} = $email;
        if ($gui->get_object('login_check_button')->get_active) {
            $CONFIG{auth} = $auth;
        }
        set_auth_key($auth);
        set_lwp_useragent();
    }
    else {
        die "Unable to login: $resp->{_content}\n";
    }
}

# ------------------ Showing/Hidding windows ------------------ #

# Main window
sub maximize_unmaximize_mainw {
    state $maximized = 0;
    $maximized++ % 2
      ? $mainw->unmaximize
      : $mainw->maximize;
}

# Users list window
sub show_users_list_window {
    $users_list_window->show;
    return 1;
}

sub hide_users_list_window {
    $users_list_window->hide;
    return 1;
}

# Help window
sub show_help_window {
    $help_window->show;
    return 1;
}

sub hide_help_window {
    $help_window->hide;
    return 1;
}

# About Window
sub show_about_window {
    $about_window->set_program_name("$appname $version");
    $about_window->set_logo($app_icon_pixbuf);
    $about_window->set_resizable(1);
    $about_window->show;
    return 1;
}

sub hide_about_window {
    $about_window->hide;
    return 1;
}

# Error window
sub hide_errors_window {
    $errors_window->hide;
    return 1;
}

# Login window
sub show_login_to_youtube_window {
    if (exists $CONFIG{email}) {
        $gui->get_object('email')->set_text($CONFIG{email});
    }
    $login_to_youtube->show;
}

sub hide_login_to_youtube_window {
    $login_to_youtube->hide;
    return 1;
}

# Details window
sub show_details_window {
    $treeview->get_selection->get_selected or return 0;
    $details_window->show;
    set_video_details();
    return 1;
}

sub hide_details_window {
    $details_window->hide;
    return 1;
}

# Feeds window
sub show_feeds_window {
    $treeview->get_selection->get_selected or return;

    $feeds_window->show;
    $feeds_statusbar->pop(0);

    if ($CONFIG{show_comments_by_default}) {
        set_comments(1);
    }
    else {
        my $iter = $feeds_liststore->append;
        $feeds_liststore->set($iter, 1, 'SHOW_COMMENTS');
        $feeds_liststore->set($iter, 0, '<big><b>Show comments for this video...</b></big>');
    }
    return 1;
}

sub hide_feeds_window {
    $feeds_liststore->clear;
    $feeds_window->hide;
    return 1;
}

# Preferences window
sub show_preferences_window {
    get_main_window_size();
    write_config_to_file();
    my $config_view_buffer = $config_view->get_buffer;
    $config_view_buffer->set_text(Config::get_config($config_file));
    $config_view->set_buffer($config_view_buffer);
    $prefernces_window->show;
    return 1;
}

sub hide_preferences_window {
    $prefernces_window->hide;
    return 1;
}

# Save plaintext config to file
sub save_configuration {
    my $config = get_text($config_view);
    Config::save_plain($config_file, $config);
    apply_configuration();
    hide_preferences_window();
    return 1;
}

# Combo boxes changes
sub combobox_sort_by_changed {
    $CONFIG{active_sort_by_combobox} = $sort_combobox->get_active;
    $CONFIG{order_by}                = $sort_combobox->get_active_text;
}

sub combobox_time_sort_by_changed {
    $CONFIG{active_time_sort_by_combobox} = $sort_by_time_combobox->get_active;
    $CONFIG{time_sort}                    = $sort_by_time_combobox->get_active_text;
}

sub combobox_resolution_changed {
    $CONFIG{active_resolution_combobox} = $resolution_combobox->get_active;
    $CONFIG{resolution} = substr($resolution_combobox->get_active_text, 0, -1);
}

sub combobox_videos_or_playlists_changed {
    $CONFIG{active_videos_playlists} = $videos_playlists_combobox->get_active;
    $CONFIG{search_playlists} = $videos_playlists_combobox->get_active_text eq 'playlists' ? 1 : 0;
}

sub combobox_safesearch_changed {
    $CONFIG{active_safe_search_combobox} = $safesearch_combobox->get_active;
    $CONFIG{safe_search}                 = $safesearch_combobox->get_active_text;
}

sub combobox_duration_changed {
    my $text = $duration_combobox->get_active_text;
    $CONFIG{active_duration_combobox} = $duration_combobox->get_active;
    $CONFIG{duration} = $text eq 'default' ? undef : $text;
}

sub combobox_caption_changed {
    my $text = $caption_combobox->get_active_text;
    $CONFIG{active_caption_combobox} = $caption_combobox->get_active;
    $CONFIG{only_videos_with_caption} = $text eq 'default' ? undef : $text;
}

sub combobox_hd_changed {
    my $text = $hd_combobox->get_active_text;
    $CONFIG{active_hd_combobox} = $hd_combobox->get_active;
    $CONFIG{only_hd_videos} = $text eq 'default' ? undef : $text;
}

# Spin buttons changes
sub spin_results_per_page_changed {
    $CONFIG{results} = $spin_results->get_value;
}

sub spin_start_with_page_changed {
    $CONFIG{start_with_page} = $spin_start_with_page->get_value;
    $CONFIG{start_index}     = $CONFIG{results} * $CONFIG{start_with_page} - $CONFIG{results} + 1;
}

# Menu items
sub toggled_mplayer_fullscreen {
    $CONFIG{fullscreen} = $fullscreen_checkbutton->get_active || 0;
}

# Check buttons toggles
sub thumbs_checkbutton_toggled {
    $CONFIG{show_thumbs} = ($_[0]->get_active || 0);
    $thumbs_column->set_visible($CONFIG{show_thumbs});
}

# Get main window size
sub get_main_window_size {
    $CONFIG{mainw_size} = join('x', $mainw->get_size);
}

my $cat_treeview   = $gui->get_object('treeview3');
my $cats_liststore = $gui->get_object('liststore4');

my $categories = [['Film &amp; Animation'      => 'Film'],
                  ['Autos &amp; Vehicles'      => 'Autos'],
                  ['Music'                     => 'Music'],
                  ['Pets &amp; Animals'        => 'Animals'],
                  ['Sports'                    => 'Sports'],
                  ['Travel &amp; Events'       => 'Travel'],
                  ['Gaming'                    => 'Games'],
                  ['Comedy'                    => 'Comedy'],
                  ['People &amp; Blogs'        => 'People'],
                  ['News &amp; Politics'       => 'News'],
                  ['Entertainment'             => 'Entertainment'],
                  ['Education'                 => 'Education'],
                  ['Howto &amp; Style'         => 'Howto'],
                  ['Nonprofits &amp; Activism' => 'Nonprofit'],
                  ['Science &amp; Technology'  => 'Tech'],
                  ['Movies'                    => 'Movies'],
                  ['Trailers'                  => 'Trailers'],
                 ];

{
    my $iter = $cats_liststore->append;
    $cats_liststore->set($iter, 0, "<big><b>\tCategories</b></big>");
}

my $i = 0;
while (defined $categories->[$i][0]) {
    my ($key, $value) = @{$categories->[$i]};
    my $iter = $cats_liststore->append;
    $cats_liststore->set($iter, 0, $key);
    $cats_liststore->set($iter, 1, "$CONFIG{feeds_main_url}/videos?category=$value");
    $cats_liststore->set($iter, 2, $feed_icon_pixbuf);
    ++$i;
}

my $tops_liststore = $gui->get_object('liststore6');
my $tops_treeview  = $gui->get_object('treeview4');

sub add_top_row {
    my ($top_url, $top_name) = @_;
    my $top_label = ucfirst $top_name;
    $top_label =~ tr/_/ /;
    my $iter = $tops_liststore->append;
    $tops_liststore->set($iter, 0, $top_label);
    $tops_liststore->set($iter, 1, $feed_icon_pixbuf);
    $tops_liststore->set($iter, 2, $top_url);
}

sub set_youtube_tops {
    my ($today, $main_label) = @_;
    my $iter = $tops_liststore->append;
    $tops_liststore->set($iter, 0, "<big><b>\t$main_label</b></big>");
    foreach my $top_name (
                          'top_rated',   'top_favorites',  'most_viewed',    'most_popular',
                          'most_recent', 'most_discussed', 'most_responded', 'recently_featured'
      ) {
        my $top_url = $CONFIG{feeds_main_url} . "/standardfeeds/$top_name";

        if ($top_name =~ /recent/) {
            next unless $today;
        }
        elsif ($today) {
            $top_url .= '?time=today';
        }
        add_top_row($top_url, $top_name);
    }
}
set_youtube_tops(1, 'Today tops');
set_youtube_tops(0, 'All time tops');

{
    my $iter = $tops_liststore->append;
    $tops_liststore->set($iter, 0, "<big><b>\tMovies</b></big>");
    foreach my $movie_top ('most_popular', 'most_recent', 'trending') {
        add_top_row($CONFIG{feeds_main_url} . '/charts/movies/' . $movie_top, $movie_top);
    }
}

# ------------ Usernames list window ------------ #
sub set_usernames {
    if (-e $CONFIG{users_list}) {
        open my $fh, '<', $CONFIG{users_list};
        while (defined(my $user = <$fh>)) {
            chomp $user;
            $users_table{lc $user} = $user;
        }
        close $fh;
    }
    foreach my $user (sort { lc $a cmp lc $b } values %users_table) {
        my $iter = $users_liststore->append;
        $users_liststore->set($iter, 0, $user);
        $users_liststore->set($iter, 1, $user_icon_pixbuf);
    }
}

sub add_username {
    my $user = $gui->get_object('username_entry')->get_text;
    $users_table{lc $user} = $user;
    my $iter = $users_liststore->append;
    $users_liststore->set($iter, 0, $user);
    $users_liststore->set($iter, 1, $user_icon_pixbuf);
}

sub add_user_to_favorites {
    my $user = get_username_from_info() or return;
    $gui->get_object('username_entry')->set_text($user);
    add_username();
    $feeds_statusbar->push(0, "Successfully added '${user}' into the username list (see: File->Users)");
}

sub get_username_from_list {
    hide_users_list_window();
    my $iter = $users_treeview->get_selection->get_selected;
    return 0 unless defined $iter;
    my $username = $users_liststore->get($iter, 0);
    videos_from_username($username);
}

sub remove_selected_user {
    my $iter = $users_treeview->get_selection->get_selected;
    my $selected_user = $users_liststore->get($iter, 0);
    delete $users_table{lc $selected_user};
    $users_liststore->remove($iter);
}

sub save_usernames_to_file {
    open my $fh, '>', $CONFIG{users_list} or return 0;
    local $, = $/;
    print {$fh} sort { lc $a cmp lc $b } values %users_table;
    close $fh;
}

# ----- My panel settings ----- #
sub log_out {
    change_subscription_page(0);
    delete $lwp_header{Authorization};
    delete $CONFIG{auth};
    $statusbar->push(1, "Not logged");
    set_lwp_useragent();
}

sub change_subscription_page {
    my ($value) = @_;
    foreach my $object (qw(subsc_scrollwindow subsc_label)) {
        $value
          ? $gui->get_object($object)->show
          : $gui->get_object($object)->hide;
    }
    return 1;
}

sub get_new_subsc {
    my $user = $gui->get_object('news_users')->get_text;
    parse_url("$CONFIG{feeds_main_url}/users/$user/newsubscriptionvideos");
}

sub get_recommended_videos {
    my $user = $gui->get_object('news_users')->get_text;
    parse_url("$CONFIG{feeds_main_url}/users/$user/recommendations");
}

sub get_favorited_videos {
    my $user = $gui->get_object('news_users')->get_text;
    parse_url("$CONFIG{feeds_main_url}/users/$user/favorites");
}

sub get_watch_history {
    my $user = $gui->get_object('news_users')->get_text;
    parse_url("$CONFIG{feeds_main_url}/users/$user/watch_history");
}

#---------------------------------#

# Returns a list of keywords (from string ref)
sub get_keywords_from_string {
    return grep { chr ord ne q{-} } split(/\s+/, ${$_[0]});
}

sub get_selected_video_code {
    my $iter = $treeview->get_selection->get_selected or return;
    my $code = $liststore->get($iter, 3);
    return wantarray ? ($code, $iter) : $code;
}

# Check if keywords are actually something else
sub check_keywords {
    given ($_[0]) {
        when (/$get_youtube_code/o) {
            get_youtube($1);
            return 1;
        }
        when (/$valid_playlist_code/o) {
            list_playlist($1);
            return 1;
        }
        when (/$get_playlist_code/o) {
            list_playlist($1);
            return 1;
        }
        when (\&is_code) {
            get_youtube($_);
            return 1;
        }
        when (m{^https?://gdata[.]youtube[.]com/feeds/}) {
            parse_url($_, (/&/ ? 1 : 0));
            return 1;
        }
        when (/$get_username/o) {
            videos_from_username($1);
            return 1;
        }
        when (m{^https?://} and /$valid_url/o) {
            set_lwp_useragent() unless exists $INC{'LWP.pm'};
            if ($lwp->get($_)->content =~ /$get_youtube_code/o) {
                get_youtube($1);
                return 1;
            }
            else {
                continue;
            }
        }
    }
    return 0;
}

sub clear_default_text {
    if ($_[0]->get_text eq 'Insert a YouTube username...') {
        $_[0]->set_text('');
    }
}

sub get_defined_pairs_from_array {

    # ('arg', undef, 'text', 'value')
    # --to--
    # ('text', 'value')

    foreach (my $i = 0 ; $i <= $#_ ; ++$i) {
        if (not defined $_[$i]) {
            splice(@_, --$i, 2);
        }
    }
    @_;
}

sub array_to_gdata_arguments {
    my @options = &get_defined_pairs_from_array;

    # ('arg', 'value', 'option', 'true')
    # --to--
    # 'arg=value&option=true'

    my $i   = -2;
    my $x   = $#options - 1;
    my $str = '';

    while (1) {
        if ($i + 3 < $x) {
            $str .= $options[$i += 2] . '=' . $options[$i + 1] . '&';
        }
        else {
            return
              $str .=
                $i + 3 == $x ? $options[-3] . '=' . $options[-2] . '&' . $options[-1]
              : $i + 2 == $x ? $options[-2] . '=' . $options[-1]
              :                $options[-1];
        }
    }
}

sub default_gdata_arguments {
    array_to_gdata_arguments(
                             'max-results' => $CONFIG{results},
                             'start-index' => $CONFIG{start_index},
                             'v'           => $constant{gdata_version},
                            );
}

sub search {
    my $keywords = $search_entry->get_text();

    return if check_keywords($keywords);
    combobox_videos_or_playlists_changed();
    do { search_playlists($keywords); return } if $CONFIG{search_playlists};

    my $author = $from_author_entry->get_text;
    unless ($author =~ /^\w+$/) {
        $author = undef;
    }

    # Get words which doesn't begins with a dash (-);
    $keywords = join(q{ }, get_keywords_from_string(\$keywords));
    $keywords = uri_escape_utf8($keywords);

    $youtube_gdata_url = "$CONFIG{feeds_main_url}/videos?"
      . array_to_gdata_arguments(
                                 'q'           => $keywords,
                                 'max-results' => $CONFIG{results},
                                 'time'        => $CONFIG{time_sort},
                                 'orderby'     => $CONFIG{order_by},
                                 'start-index' => $CONFIG{start_index},
                                 'safeSearch'  => $CONFIG{safe_search},
                                 'hd'          => $CONFIG{only_hd_videos},
                                 'caption'     => $CONFIG{only_videos_with_caption},
                                 'duration'    => $CONFIG{duration},
                                 'author'      => $author,
                                 'v'           => $constant{gdata_version}
                                );

    if (defined $CONFIG{query_parameters}) {
        unless ($CONFIG{query_parameters} =~ /^&/) {
            substr($CONFIG{query_parameters}, 0, 0, '&');
        }
        $youtube_gdata_url .= $CONFIG{query_parameters};
    }
    parse_content($youtube_gdata_url);
    print_results();
}

#---------------------- PREPARE GDATA FEEDS URL ----------------------#
sub parse_url {
    $youtube_gdata_url = shift;
    my $as_it_is = @_ ? shift() : 0;

    unless ($as_it_is) {
        $youtube_gdata_url .= $youtube_gdata_url =~ /\?/ ? '&' : '?';
        $youtube_gdata_url .= default_gdata_arguments();
    }
    $videos_playlists_combobox->set_active(0);
    parse_content($youtube_gdata_url);
    print_results();
}

#---------------------- GET AND PARSE GDATA CONTENT ----------------------#
sub parse_content {

    undef @results;
    my $number = 0;
    my $hash   = xml2hash(lwp_get($_[0]));

    while (
           my $gdata =
             ref $hash->{'feed'}{'entry'} eq 'ARRAY' ? $hash->{'feed'}{'entry'}[$number]
           : ref $hash->{'feed'}{'entry'} eq 'HASH'  ? $hash->{'feed'}{'entry'}
           : $hash->{'entry'}
      ) {
        last unless defined $gdata;

        push @results, $CONFIG{search_playlists}

          # Playlists
          ? {
             'playlistID' => $gdata->{'yt:playlistId'},
             'title'      => $gdata->{'title'},
             'author'     => $gdata->{'author'}{'name'},
             'count'      => $gdata->{'yt:countHint'},
             'summary'    => $gdata->{'summary'},
             'published'  => $gdata->{'published'},
             'updated'    => $gdata->{'updated'},
            }

          # Videos
          : {
             'code'        => $gdata->{'media:group'}{'yt:videoid'},
             'title'       => $gdata->{'media:group'}{'media:title'}{'#text'},
             'author'      => $gdata->{'author'}{'name'},
             'rating'      => $gdata->{'gd:rating'}{'-average'},
             'likes'       => $gdata->{'yt:rating'}{'-numLikes'} || 0,
             'dislikes'    => $gdata->{'yt:rating'}{'-numDislikes'} || 0,
             'favorited'   => $gdata->{'yt:statistics'}{'-favoriteCount'},
             'duration'    => format_time($gdata->{'media:group'}{'yt:duration'}{'-seconds'}),
             'views'       => $gdata->{'yt:statistics'}{'-viewCount'},
             'published'   => $gdata->{'published'},
             'description' => $gdata->{'media:group'}{'media:description'}{'#text'},
             'category'    => ref $gdata->{'category'} eq 'ARRAY' ? $gdata->{'category'}[1]{'-label'}
             : $gdata->{'category'}{'-label'} || 'Unknown',
            };
        ++$number;
        last unless ref $hash->{'feed'}{'entry'} eq 'ARRAY';
    }

    die "No results found!\n\n", Config::_dump($hash), "\n" unless @results;
    return 1;
}

# Format seconds to HH:MM:SS
sub format_time {
    $_[0] >= 3600
      ? join ':', map { sprintf '%02d', $_ } $_[0] / 3600 % 24, $_[0] / 60 % 60, $_[0] % 60
      : join ':', map { sprintf '%02d', $_ } $_[0] / 60 % 60, $_[0] % 60;
}

# Thousand separator
sub set_thousands {
    return 0 unless $_[0];
    length($_[0]) > 3 or return $_[0];
    my $n = shift;
    my $l = length($n) - 3;
    my $i = ($l - 1) % 3 + 1;
    my $x = substr($n, 0, $i) . $CONFIG{thousand_separator};
    while ($i < $l) {
        $x .= substr($n, $i, 3) . $CONFIG{thousand_separator};
        $i += 3;
    }
    $x . substr($n, $i);
}

sub format_date {
    $_[0] =~ m{^(\d{4})-(\d{2})-(\d{2})} ? "$2/$3/$1" : $_[0];
}

#---------------------- PRINT VIDEO RESULTS ----------------------#
sub encode_entities {
    my ($text) = @_;
    $text =~ s/&/&amp;/g;
    $text =~ s/</&lt;/g;
    $text =~ s/>/&gt;/g;
    return $text;
}

sub get_next_page_spaces {
    $CONFIG{show_thumbs} ? q{ } x 50 : q{ } x 100;
}

sub append_next_page {
    my $iter = $liststore->append;
    $liststore->set($iter, 0, get_next_page_spaces() . "<big><b>NEXT PAGE</b></big>");
}

sub get_code {
    my ($code, $iter) = get_selected_video_code() or return;

    defined $code
      ? $code =~ /$valid_playlist_code/o
          ? do { $CONFIG{search_playlists} = 0; list_playlist($code) }
          : get_youtube($code)
      : do {
        $liststore->set(
            $iter, 0,
            get_next_page_spaces() . do {
                $youtube_gdata_url =~ /[&?]start-index=(\d+)/;
                '<big><b>Page: ' . ((($1 - 1) / $CONFIG{results}) + 2) . '</b></big>';
              }
        );
        next_page();
      };
}

sub print_results {
    $liststore->clear if $CONFIG{clear_list} and not @_;

    my $num = 0;
    foreach my $video (@results) {
        ++$num;
        my $iter = $liststore->append;
        my $row_description = $video->{description} // 'No description available...';
        $liststore->set($iter, 4, $row_description);
        $row_description = join(' ', split(' ', $row_description));
        $row_description =~ s/(.)\1{3,}/$1/sg;
        $liststore->set($iter, 3, $video->{code});
        $liststore->set(
                        $iter,
                        0,
                        "<big><b>"
                          . encode_entities($video->{title})
                          . "</b></big>\n\n"
                          . "<b>Likes\t:</b> "
                          . set_thousands($video->{likes}) . "\n"
                          . "<b>Dislikes\t:</b> "
                          . set_thousands($video->{dislikes}) . "\n"
                          . "<b>Favorited\t:</b> "
                          . set_thousands($video->{favorited}) . "\n"
                          . "<b>Category\t:</b> "
                          . encode_entities($video->{category}) . "\n"
                          . "<b>Uploader\t: </b>"
                          . $video->{author} . "\n\n" . "<i>"
                          . encode_entities($row_description) . "</i>"
                       );

        $liststore->set(
            $iter, 2,
            "<b>Length\t:</b> "
              . $video->{duration} . "\n"
              . "<b>Rating\t:</b> "
              . sprintf('%.2f', ($video->{rating} // 0))
              . "\n"    # /
              . "<b>Views\t:</b> "
              . set_thousands($video->{views})
              . do {
                defined $video->{published}
                  ? "\n" . "<b>Added\t: </b>" . format_date($video->{published})
                  : '';
              }
        );

        if ($CONFIG{show_thumbs}) {
            my $pixbufloader = 'Gtk2::Gdk::PixbufLoader'->new;
            $pixbufloader->set_size(120, 90);
            if ($CONFIG{use_threads_for_thumbs}) {
                $pixbufloader->write(threads_lwp_get(sprintf($CONFIG{thumb_url}, $video->{code}, $CONFIG{default_thumb})));
            }
            else {
                $pixbufloader->write($lwp->get(sprintf($CONFIG{thumb_url}, $video->{code}, $CONFIG{default_thumb}))->content);
            }
            my $pixbuf = $pixbufloader->get_pixbuf;
            $pixbufloader->close;
            $pixbuf = $default_thumb unless $pixbuf;
            $liststore->set_value($iter, 1, $pixbuf);
        }
    }
    append_next_page() if @results;
}

#---------------------- PRINT PLAYLISTS ----------------------#
sub print_playlists {
    my ($playlist) = @_;

    my $iter = $liststore->append;
    my $row_description = $playlist->{summary} || 'No description available...';
    $liststore->set($iter, 4, $row_description);
    $row_description = join(' ', split(' ', $row_description));
    $row_description =~ s/(.)\1{3,}/$1/sg;

    $liststore->set($iter, 3, $playlist->{playlistID});
    $liststore->set(
                    $iter,
                    0,
                    '<big><b>'
                      . encode_entities($playlist->{'title'})
                      . "</b></big>\n\n"
                      . "<b>Author\t:</b> "
                      . $playlist->{author} . "\n"
                      . "<b>Published\t:</b> "
                      . format_date($playlist->{published}) . "\n"
                      . "<b>Updated\t:</b> "
                      . format_date($playlist->{updated}) . "\n\n" . '<i>'
                      . encode_entities($row_description) . '</i>'
                   );

    $liststore->set($iter, 2, "<b>Videos:</b> " . set_thousands($playlist->{count}) . "\n\n");

    if ($CONFIG{show_thumbs}) {
        $liststore->set_value($iter, 1, $default_thumb);
    }
}

sub list_playlist {
    parse_url("$CONFIG{feeds_main_url}/playlists/$_[0]");
}

# Get playlists from username
sub playlists_from_selected_username {
    my $iter = $users_treeview->get_selection->get_selected;
    playlists_from_username($users_liststore->get($iter, 0));
}

sub videos_from_text_entry_username {
    videos_from_username($_[0]->get_text);
}

# Get videos from username
sub videos_from_username {
    my $iter = $users_treeview->get_selection->get_selected;
    my $username = ref $_[0] ? $users_liststore->get($iter, 0) : $_[0];
    parse_url("$CONFIG{feeds_main_url}/users/$username/uploads");
}

sub playlists_from_username {
    my $username = ref $_[0] ? $_[0]->get_text : $_[0];
    $videos_playlists_combobox->set_active(1);
    search_playlists("$CONFIG{feeds_main_url}/users/$username/playlists?" . default_gdata_arguments());
}

sub search_playlists {
    my ($arg, $next_page) = @_;
    if (not $next_page and $CONFIG{clear_list}) {
        $liststore->clear;
    }
    $youtube_gdata_url =
        $arg =~ /^\Q$CONFIG{feeds_main_url}\E/o
      ? $arg
      : $CONFIG{feeds_main_url} . "/playlists/snippets?q=$arg&" . default_gdata_arguments();

    parse_content($youtube_gdata_url);
    foreach my $playlist (@results) {
        print_playlists($playlist);
    }
    append_next_page();
}

#---------------------- NEXT PAGE ----------------------#
sub next_page {
    if ($youtube_gdata_url =~ s/[&?]start-index=\K(\d+)/($1 + $CONFIG{results})/e) {
        if ($CONFIG{search_playlists}) {
            search_playlists($youtube_gdata_url, 1);
        }
        else {
            parse_content($youtube_gdata_url);
            print_results(1);
        }
    }
}

sub select_resolution {
    given ($CONFIG{'resolution'}) {
        when (1080) {
            return $streaming{1080} // lower_quality();
        }
        when (720) {
            return $streaming{720} // lower_quality();
        }
        when (480) {
            return $streaming{480} // lower_quality();
        }
        when (360) {
            return $streaming{360} // lower_quality();
        }
        when (240) {
            return $streaming{240} // lower_quality();    #/
        }
        default {
            return lower_quality();                       # /
        }
    }
}

sub lower_quality {
    foreach my $itag (sort { $b <=> $a } values %itags) {
        if (exists($streaming{$itag})) {
            return $streaming{$itag};
        }
    }
}

sub format_itags {
    my @itags;
    foreach my $itag (@_) {
        if (exists($itags{$itag})) {
            push @itags, $itags{$itag};
        }
    }
    @itags;
}

sub check_mplayer_arguments {
    if (   $CONFIG{use_lower_cache}
        or not exists $streaming{720}
        or $CONFIG{resolution} < 720) {
        set_mplayer_arguments($CONFIG{lower_cache}, $CONFIG{lower_cache_min});
    }
    else {
        set_mplayer_arguments($CONFIG{cache}, $CONFIG{cache_min});
    }
}

#---------------------- PLAY AN YOUTUBE VIDEO ----------------------#
sub play_video {
    my ($streaming) = @_;

    # Update mplayer's arguments
    check_mplayer_arguments();

    my $mplayer_line = join(' ', map { quotemeta }
                              split(' ', join(' ', $CONFIG{mplayer}, values %MPLAYER)));

    if ($CONFIG{debug}) {
        print "\n** STREAMING: $streaming\n\n", "** MPlayer Line: $mplayer_line\n\n";
    }

    system $mplayer_line . ' ' . quotemeta($streaming) . ' &';

    # Change directory back to the main working directory
    chdir delete $constant{'cwd'} if exists $constant{'cwd'};
}

# Getting YouTube closed captions with gcap
sub get_closed_caption {
    my ($code) = @_;

    unless (exists $constant{'cwd'}) {
        $constant{'cwd'} = rel2abs(curdir());
    }

    # Change dir to $TMP and get the SRT file
    chdir $CONFIG{tmp_dir};

    my $srt_file;
    my $i = 0;
    {
        $srt_file =
          -e "${code}_$CONFIG{srt_language}.srt"
          ? "${code}_$CONFIG{srt_language}.srt"
          : do {
            opendir(my $dir_h, $CONFIG{tmp_dir}) or return q{};
            my $srt = (grep /^\Q$code\E[\w-]*[.](?i:srt)\z/, readdir($dir_h))[0];
            closedir $dir_h;
            $srt;
          };

        unless (defined $srt_file) {
            system $CONFIG{perl_bin}, $CONFIG{gcap}, $code;

            if ($? == 0 and not $i++) {
                redo;
            }
        }
    }

    return defined $srt_file
      ? "$CONFIG{mplayer_srt_settings} -sub $srt_file"
      : '';
}

sub get_youtube {
    my ($code) = @_;

    my $content = lwp_get("$CONFIG{get_video_info}?&video_id=$code&el=detailpage&ps=default&eurl=&gl=US&hl=en");
    my $url     = "http://www.youtube.com/watch?v=$code";

    $MPLAYER{arguments} = '';
    if (-e $CONFIG{gcap} and $content =~ /&has_cc=True&/) {
        $MPLAYER{arguments} = get_closed_caption($code);
    }

    if ($content =~ /url_encoded_fmt_stream_map=(.+?)&/) {
        my $streaming = $1;
        $streaming =~ s/%253A/:/gi;
        $streaming =~ s{%252F}{/}gi;
        $streaming =~ s/%2526/&/g;
        $streaming =~ s/%253D/=/gi;
        $streaming =~ s/%253F/?/gi;
        $streaming =~ s/%25252C/,/gi;

        undef %streaming;
        my (@streaming_urls) =
          grep /^https?:/, split(/url%3D(.+?)%26/, $streaming);
        @streaming{format_itags(map /&itag=(\d+)&/, @streaming_urls)} =    #/
          grep {
            exists $itags{
                do { /&itag=(\d+)&/; $1 }
              }
          } @streaming_urls;

        # Select one resolution and play video
        play_video(select_resolution());
    }
    else {

        # This happens when video has been deleted or is forbidden
        die "(x_x) Something went wrong...\n\n", "(x_x) Unable to stream: $url\n\n";
    }
}

sub list_category {
    my $iter = $cat_treeview->get_selection->get_selected;
    parse_url($cats_liststore->get($iter, 1));
}

sub list_tops {
    my $iter = $tops_treeview->get_selection->get_selected;
    parse_url($tops_liststore->get($iter, 2));
}

sub clear_text {
    $_[0]->set_text('') if $CONFIG{clean_text_entries_on_click};
    return 0;
}

sub run_cli_youtube_viewer {
    execute_external_program($CONFIG{terminal}, '-e', q{'} . join(' ', $CONFIG{perl_bin}, $CONFIG{cli_youtube_viewer}), q{'});
}

sub find_terminal {
    foreach my $term (
                      'gnome-terminal', 'lxterminal', 'terminal', 'xfce4-terminal',
                      'sakura',         'lilyterm',   'evilvte',  'superterm',
                      'terminator',     'kterm',      'mlterm',   'mrxvt',
                      'rxvt',           'urxvt',      'uuterm',   'termit',
                      'fbterm',         'stjerm',     'yakuake',  'roxterm'
      ) {
        foreach my $path (@path) {
            if (-e "$path/$term") {
                return "$path/$term";
            }
        }
    }
    return ($ENV{TERM} // 'xterm');    # /
}

sub get_options_as_arguments {
    my @args;
    my %options = (
                   'order-by'     => $CONFIG{order_by},
                   'time'         => $CONFIG{time_sort},
                   'sub'          => $CONFIG{srt_language},
                   'download-dir' => $CONFIG{downloads_folder},
                   'lower-cache'  => $CONFIG{use_lower_cache} ? 0 : undef,
                   '-fullscreen'  => $CONFIG{fullscreen} ? 0 : undef,
                  );
    $options{$CONFIG{resolution} . 'p'} = 0;
    while (my ($argv, $value) = each %options) {
        push(
            @args,
            do {
                $value             ? '-' . $argv . '=' . $value
                  : defined $value ? '-' . $argv
                  :                  next;
              }
            );
    }
    return @args;
}

sub is_code {
    length $_[0] == 11
      and $_[0] =~ /^[-_0-9A-Za-z]{11}\z/
      and $_[0] =~ /[-_0-9A-Z]/;
}

sub execute_external_program {
    system join(' ', map { chr ord eq q{'} ? $_ : quotemeta } @_) . ' &';
}

sub _make_youtube_url {
    sprintf(
              $CONFIG{search_playlists}
            ? $CONFIG{playlists_url}
            : $CONFIG{youtube_video_url}, $_[0]
           );
}

sub open_youtube_url {
    my $url = _make_youtube_url(get_selected_video_code() or return);
    execute_external_program((defined $CONFIG{browser} ? $CONFIG{browser} : 'xdg-open'), $url);
    return 1;
}

my @queue_codes;

sub queue_playback {
    my $code = get_selected_video_code() or return;
    print "[*] Added: $code\n" if $CONFIG{debug};
    push @queue_codes, $code;
    return 1;
}

sub play_all_video_results {
    execute_cli_youtube_viewer([map { $_->{code} } @results]);
    return 1;
}

sub play_selected_video_with_cli_youtube_viewer {
    my $code = get_selected_video_code() or return;
    execute_cli_youtube_viewer([$code]);
    return 1;
}

sub play_videos_from_queue {
    if (@queue_codes) {
        execute_cli_youtube_viewer(\@queue_codes);
        undef @queue_codes;
    }
    return 1;
}

sub execute_cli_youtube_viewer {
    my $code_array_ref = shift();

    # Executing youtube-viewer
    system join(
                ' ',
                (
                 map { chr ord eq q{'} ? $_ : quotemeta } $CONFIG{terminal},
                 '-e',
                 q{'}
                   . join(' ',
                          $CONFIG{perl_bin},
                          $CONFIG{cli_youtube_viewer},
                          (map { sprintf($CONFIG{youtube_video_url}, $_) } @{$code_array_ref}),
                          get_options_as_arguments(), @_)
                   . q{'}
                )
               ) . ' &';
    return 1;
}

sub download_video {
    my $code = get_selected_video_code() or return;
    die "Unable to download a playlist!\n" if $CONFIG{search_playlists};
    execute_cli_youtube_viewer([$code], '--download');
    return 1;
}

sub comments_row_activated {
    my $iter = $feeds_treeview->get_selection->get_selected() or return;
    my $value = $feeds_liststore->get($iter, 1);

    if ($value eq 'SHOW_COMMENTS') {
        set_comments(1);
    }
    elsif ($value =~ /^[0-9]+\z/) {
        set_comments($value + $constant{comments_per_page});
    }
    else {
        execute_external_program((defined $CONFIG{browser} ? $CONFIG{browser} : 'xdg-open'), $value);
    }
}

sub get_username_from_info {
    my $iter = $treeview->get_selection->get_selected() or return;
    my ($username) = $liststore->get($iter, 0) =~ /\bUploader\s*:\s*<.+?>\s*(\S+)/i;
    return $username;
}

sub goto_youtube_user_profile {
    my $url = sprintf($CONFIG{youtube_profile_url}, get_username_from_info());
    execute_external_program((defined $CONFIG{browser} ? $CONFIG{browser} : 'xdg-open'), $url);
}

sub show_related_videos {
    my $code = get_selected_video_code() or return;
    parse_url("$CONFIG{feeds_main_url}/videos/$code/related");
}

sub send_rating_to_video {
    my ($rating) = @_;
    my $code = get_selected_video_code() or return;
    my $uri = "$CONFIG{feeds_main_url}/videos/$code/ratings";

    _save(
        'POST', $uri, <<"XML_HEADER"
<?xml version="1.0" encoding="UTF-8"?>
<entry xmlns="http://www.w3.org/2005/Atom"
       xmlns:yt="http://gdata.youtube.com/schemas/2007">
<yt:rating value="$rating"/>
</entry>
XML_HEADER
         );
}

sub send_comment_to_video {
    my $comment = get_text($gui->get_object('comment_textview'));

    return unless length $comment;

    my $code = get_selected_video_code() or return;
    my $uri = "$CONFIG{feeds_main_url}/videos/$code/comments";

    $feeds_statusbar->push(
        0,
        _save(
            'POST', $uri, <<"XML_HEADER"
<?xml version="1.0" encoding="UTF-8"?>
<entry xmlns="http://www.w3.org/2005/Atom"
    xmlns:yt="http://gdata.youtube.com/schemas/2007">
  <content>$comment</content>
</entry>
XML_HEADER
          ) ? 'Comment sent' : 'Error!'
    );
}

sub subscribe_channel {
    my $user = get_username_from_info() or return;
    my $uri = "$CONFIG{feeds_main_url}/users/default/subscriptions";

    $feeds_statusbar->push(
        0,
        _save(
            'POST', $uri, <<"XML_HEADER"
<?xml version="1.0" encoding="UTF-8"?>
<entry xmlns="http://www.w3.org/2005/Atom"
  xmlns:yt="http://gdata.youtube.com/schemas/2007">
    <category scheme="http://gdata.youtube.com/schemas/2007/subscriptiontypes.cat"
      term="channel"/>
    <yt:username>$user</yt:username>
</entry>
XML_HEADER
          ) ? "Successfully subscribed to user: $user" : 'Error!'
    );
}

sub favorite_video {
    my $code = get_selected_video_code() or return;
    my $uri = "$CONFIG{feeds_main_url}/users/default/favorites";

    $feeds_statusbar->push(
        0,
        _save(
            'POST', $uri, <<"XML_HEADER"
<?xml version="1.0" encoding="UTF-8"?>
<entry xmlns="http://www.w3.org/2005/Atom">
  <id>$code</id>
</entry>
XML_HEADER
          ) ? "Successfully favorited video: $code" : 'Error!'
    );
}

sub _request {
    my ($req) = @_;

    my $res = $lwp->request($req);
    if ($res->is_success) {
        return $res->content();
    }
    else {
        die "Error: $res->code\n", $res->content;
    }
}

sub _prepare_request {
    my ($req, $length) = @_;
    $req->header('GData-Version' => 2);
    $req->header('Content-Length' => $length) if ($length);

    if (defined $lwp_header{'Authorization'}) {
        $req->header('Authorization' => $lwp_header{'Authorization'});
        $req->header('X-GData-Key'   => $lwp_header{'X-GData-Key'});
    }
}

sub _save {
    my ($method, $uri, $content) = @_;

    my $req = HTTP::Request->new("$method" => $uri);
    $req->content_type('application/atom+xml; charset=UTF-8');
    _prepare_request($req, length($content));
    $req->content($content);

    return _request($req);
}

sub like_selected_video {
    $feeds_statusbar->push(0, send_rating_to_video('like') ? 'Video liked' : 'Error!');
}

sub dislike_selected_video {
    $feeds_statusbar->push(0, send_rating_to_video('dislike') ? 'Video disliked' : 'Error!');
}

sub set_comments {
    my $index = shift;
    my $code = get_selected_video_code() or return;
    $feeds_liststore->clear if $index == 1;

    my $hash = xml2hash(
                        lwp_get(
                                "$CONFIG{feeds_main_url}/videos/$code/comments?"
                                  . array_to_gdata_arguments(
                                                             'v'           => $constant{gdata_version},
                                                             'start-index' => $index
                                                            )
                               )
                       );

    my $number = 0;
    while (
           my $gdata =
             ref $hash->{'feed'}{'entry'} eq 'ARRAY' ? $hash->{'feed'}{'entry'}[$number]
           : ref $hash->{'feed'}{'entry'} eq 'HASH'  ? $hash->{'feed'}{'entry'}
           : $hash->{'entry'}
      ) {
        last unless defined $gdata;

        my $iter = $feeds_liststore->append;
        $feeds_liststore->set($iter, 1, sprintf($CONFIG{youtube_profile_url}, $gdata->{author}{name}));
        $feeds_liststore->set(
                              $iter,
                              0,
                              "<big><b>$gdata->{author}{name}</b> ("
                                . format_date($gdata->{updated})
                                . ") said:</big>\n\t"
                                . encode_entities($gdata->{content})
                             );

        ++$number;
        last unless ref $hash->{'feed'}{'entry'} eq 'ARRAY';
    }

    if ($number) {
        my $iter = $feeds_liststore->append;
        $feeds_liststore->set($iter, 0, "\n<big><b>=&gt;&gt; NEXT PAGE</b></big>\n");
        $feeds_liststore->set($iter, 1, $index);
    }

    return 1;
}

sub show_more_videos_from_username {
    videos_from_username(get_username_from_info() or return);
}

# Setting details to details_window
sub set_video_details {
    my ($code, $iter) = get_selected_video_code();
    my $main_details = $liststore->get($iter, 0);

    my %labels = (
                  Added  => 'Published',
                  Length => 'Duration',
                 );

    # Setting title
    my $title = substr($main_details, 0, index($main_details, '</big>') + 6, '');
    $gui->get_object('video_title_label')->set_label("<big>$title</big>");

    # Setting video details
    $main_details =~ s/^\s+//;
    $main_details =~ s{\s*<i>.+</i>\s*}{\n};
    $main_details =~ s/\s+/$-[0] < 10 ? "\t\t" : "\t"/e;
    my $secondary_details = $liststore->get($iter, 2);
    $secondary_details =~ s/:(?!\d)/\t: /g;
    $secondary_details =~ s/\b$_\b\t/$labels{$_}/ for keys %labels;
    $gui->get_object('video_details_label')->set_label($main_details . $secondary_details);

    # Setting the link button
    my $url        = _make_youtube_url($code);
    my $linkbutton = $gui->get_object('linkbutton1');
    $linkbutton->set_label($url);
    $linkbutton->set_uri($url);

    if ($CONFIG{search_playlists}) {
        foreach my $nr (qw(1 2 3)) {
            $gui->get_object("image$nr")->set_from_pixbuf($default_thumb);
        }
    }
    else {

        # Getting thumbs
        foreach my $nr (qw(1 2 3)) {
            my $pixbufloader = 'Gtk2::Gdk::PixbufLoader'->new;
            if ($CONFIG{use_threads_for_thumbs}) {
                $pixbufloader->write(threads_lwp_get(sprintf($CONFIG{thumb_url}, $code, $nr)));
            }
            else {
                $pixbufloader->write($lwp->get(sprintf($CONFIG{thumb_url}, $code, $nr))->content);
            }
            $pixbufloader->close;
            $gui->get_object("image$nr")->set_from_pixbuf($pixbufloader->get_pixbuf);
        }
    }

    # Setting textview description
    set_text($gui->get_object('description_textview'), $liststore->get($iter, 4));
    return 1;
}

sub on_mainw_destroy {

    # Save hpaned position
    $CONFIG{hpaned_position} = $hbox2->get_position;

    get_main_window_size();
    write_config_to_file();
    save_usernames_to_file();
    'Gtk2'->main_quit;
    exit 0;
}

$notebook->set_current_page($CONFIG{default_notebook_page});

'Gtk2'->main;
